/*
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *  HTTP/1.1 compliance by Arkadiusz Miskiewicz <misiek@pld.org.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "others.h"
#include "options.h"
#include "defaults.h"
#include "manager.h"
#include "shared.h"

int downloadmanager(const struct optstruct *opt)
{
	char *newdir, *oldmd5 = NULL, *newmd5 = NULL, *downmd5, *tempname;
	int nodb = 0;
	int hostfd;
	time_t currtime;


    time(&currtime);
    logg("Checking for new database - started at %s", ctime(&currtime));
    mprintf("Checking for new database - started at %s", ctime(&currtime));

    /* change current working directory to selected */
    if(optl(opt, "datadir"))
	newdir = getargl(opt, "datadir");
    else
	newdir = VIRUSDBDIR;

    if(chdir(newdir)) {
	mprintf("Can't change dir to %s\n", newdir);
	return 50;
    }

    /* first thing we want, is a local file md5 checksum */
    if(info("viruses.db", 1) == -1) {
	/* there is no database, so we just download new one */
	nodb = 1; 
	mprintf("Database not found in the local directory.\n");
    } else {
	if((oldmd5 = md5_file("viruses.db")) == NULL) {
	    mprintf("@Can't create md5 checksum of virus database.\n");
	    return 51;
	}
    }

    /*
     * Ok, we have local file md5 checksum, now we download md5sum of current
     * database from Internet.
     */

    hostfd = wwwconnect(HOST);

    if(hostfd < 0) {
	mprintf("@Connection with "HOST" failed.\n");
	return 52;
    }


    if((newmd5 = get_md5_checksum(hostfd, HOST)) == NULL) {
	mprintf("@Can't get md5 sum from "HOST"\n");
	close(hostfd);
	return 52;
    }

    /* ok, we have md5 sum from Internet */

    if(oldmd5 && !strncmp(oldmd5, newmd5, 32)) {
	mprintf("Database is up to date.\n");
	logg("Database is up to date.\n");
	close(hostfd);
	return 1;
    }

    /* temporary file is created in clamav's directory thus we don't need
     * to create it immediately, because race conditions are impossible
     */
    tempname = gentemp(".");

    /* download the database */

    if(get_database(hostfd, tempname, HOST)) {
	mprintf("@Can't download the database from "HOST"\n");
	close(hostfd);
	return 52;
    }

    close(hostfd);

    /* ok, we have new database */
    if((downmd5 = md5_file(tempname)) == NULL) {
	unlink(tempname);
        mprintf("@Can't create md5 checksum of virus database.\n");
        return 51;
    }

    if(!strcmp(newmd5, downmd5)) {
	if(!nodb && unlink("viruses.db")) {
	    mprintf("@Can't unlink old file. Rename the files yourself.\n");
	    free(tempname);
	    return 53;
	}
	rename(tempname, "viruses.db");
	mprintf("Database updated.\n");
	free(tempname);
	return 0;
    } else {
	mprintf("@The checksum of downloaded database isn't ok. Please check it yourself or try again.\n");
	unlink(tempname);
	free(tempname);
	return 54;
    }

}


/* this function returns socket descriptor */

int wwwconnect(char *server)
{
	int socketfd;
	struct sockaddr_in name;
	struct hostent *host;

    socketfd = socket(PF_INET, SOCK_STREAM, 0);
    name.sin_family = AF_INET;

    if((host = gethostbyname(server)) == NULL) {
	mprintf("@Can't get information about %s host.\n", server);
	return -1;
    }

    mprintf("Connecting to %s\n", server);

    name.sin_addr = *((struct in_addr *) host->h_addr);
    name.sin_port = htons(80);

    if(connect(socketfd, (struct sockaddr *) &name, sizeof(struct sockaddr_in)) == -1) {
	mprintf("@Can't connect to port 80 of host %s\n", server);
	return -2;
    }

    return socketfd;
}

char *get_md5_checksum(int socketfd, const char *hostname)
{
	char cmd[300], buffer[FBUFFSIZE];
	char *md5sum, *ch;
	int i, j, bread;

    mprintf("Reading md5 sum of database from %s : ", hostname);

    snprintf(cmd, 300, "GET "MD5FILE" HTTP/1.1\r\n"
	     "Host: %s\r\n"
	     "User-Agent: "PACKAGE"/"VERSION"\r\n"
//	     "Connection: close\r\n"
	     "\r\n", hostname);
    write(socketfd, cmd, strlen(cmd));

    if ((bread = recv(socketfd, buffer, FBUFFSIZE, 0)) == -1) {
      mprintf("@Error while reading md5 sum of database from %s\n", hostname);
      return NULL;
    }

    if ((strstr(buffer, "HTTP/1.1 404")) != NULL) { 
      mprintf("@md5 sum not found on remote server\n");
      return NULL;
    }

    ch = buffer;
    i = 0;
    while (1) {
      if (*ch == '\n' && *(ch - 1) == '\r' && *(ch - 2) == '\n' && *(ch - 3) == '\r') {
	ch++;
	i++;
	break;
      }
      ch++;
      i++;
    }  

    md5sum = (char *) mcalloc(33, sizeof(char));
    
    for (j=0; j<32; j++) {
      if (!ch || (ch && !*ch) || (ch && !isalnum(ch[j]))) {
	mprintf("@Malformed md5 checksum detected.\n");
	free(md5sum);
	return NULL;
      }
      md5sum[j] = ch[j];
    }

    mprintf("OK\n");
    return md5sum;
}

int get_database(int socketfd, const char *file, const char *hostname)
{
	char cmd[300], buffer[FBUFFSIZE];
	char *ch;
	int bread, fd, i;


    if((fd = open(file, O_WRONLY|O_CREAT|O_EXCL, 0644)) == -1) {
	mprintf("@Can't open new file %s to write\n", file);
	return -1;
    }

    snprintf(cmd, 300, "GET "DBFILE" HTTP/1.1\r\n"
	     "Host: %s\r\n"
	     "User-Agent: "PACKAGE"/"VERSION"\r\n"
	     "Connection: close\r\n"
	     "\r\n", hostname);
    write(socketfd, cmd, strlen(cmd));

    if ((bread = recv(socketfd, buffer, FBUFFSIZE, 0)) == -1) {
      mprintf("@Error while reading database from %s\n", hostname);
      return -1;
    }

    if ((strstr(buffer, "HTTP/1.1 404")) != NULL) { 
      mprintf("@Database file not found on remote server\n");
      return -1;
    }

    mprintf("Downloading database from %s ", hostname);

    ch = buffer;
    i = 0;
    while (1) {
      if (*ch == '\n' && *(ch - 1) == '\r' && *(ch - 2) == '\n' && *(ch - 3) == '\r') {
	ch++;
	i++;
	break;
      }
      ch++;
      i++;
    }  

    write(fd, ch, bread - i);
    while((bread = read(socketfd, buffer, FBUFFSIZE))) {
	write(fd, buffer, bread);
	mprintf(".");
    }

    mprintf(" done\n");
    close(fd);
    return 0;
}
